/*
 *  linux/arch/arm/kernel/setup-sa.c
 *
 *  Copyright (C) 1995, 1996 Russell King
 */

/*
 * This file obtains various parameters about the system that the kernel
 * is running on.
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/ldt.h>
#include <linux/user.h>
#include <linux/a.out.h>
#include <linux/tty.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/major.h>
#include <linux/utsname.h>

#include <asm/segment.h>
#include <asm/system.h>
#include <asm/hardware.h>
#include <asm/pgtable.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/proc/unaligned.h>
#include <asm/setup.h>

#define COMMAND_LINE_SIZE 512

unsigned long arm_id;
extern int root_mountflags;
extern int _etext, _edata, _end;

#ifdef CONFIG_BLK_DEV_RAM
extern int rd_doload;		/* 1 = load ramdisk, 0 = don't load */
extern int rd_prompt;		/* 1 = prompt for ramdisk, 0 = don't prompt */
extern int rd_image_start;	/* starting block # of image */

char Read_Byte (void);

static inline void setup_ramdisk (void)
{
    rd_image_start	= 0;
    rd_prompt		= 1;
    rd_doload		= 1;
}
#else
#define setup_ramdisk()	
#endif

static char default_command_line[COMMAND_LINE_SIZE] = "mem=16M  root=/dev/hda1 hda=autotune hdb=autotune ";
static char command_line[COMMAND_LINE_SIZE] = { 0, };
       char saved_command_line[COMMAND_LINE_SIZE];

//these variables are to keep track of shared write-only I/O registers and CPLD
static	unsigned int cShadowX = 0;//what was actually written into CPLD + sw lock
static	int semaphoreID=0;	//store here ID of a caller, to avoid reentrancy
void    tune_up_pci_clock(void);
static	void vga_init(void);
static	void setup_other_io(void);
static	void min_pci_bios(void);

unsigned int	serial_num = 0;

struct processor processor;
extern const struct processor sa110_processor_functions;

extern void drum(void);		//startup-tune playing routine
extern void drum1(void);	//shutdown-tune playing routine

unsigned char aux_device_present = 0xAA;       // yes we have a mouse
struct screen_info screen_info = {
        0, 0, { 0, 0 },        // x, y, unused
        0, 0, 80,      // page, mode. cols
        0, 0, 0,       // unused
        25, 1, 16      // lines, isVGA, points
};


void putstring(char *s)
{
  while (*s != '\0') {
    if (*s == '\n') putcharV('\r') ;
    putcharV(*s) ;
    s++ ;
  }
}

void puthex(unsigned int n)
{
  static unsigned char string[8] ;
  int i ;

  for (i = 0; i < 8; i++) {
    string[i] = (n & 0xf) ;
    n = n >> 4 ;
  }

  putstring("0x") ;
  for (i = 7; i >= 0 ; i--) {
    if (string[i] >= 0xA)
      putcharV(string[i] - 0xA + 'A') ;
    else
      putcharV(string[i] + '0') ;
  }
}

// set up the PCI devices
static void min_pci_bios(void)
{
volatile unsigned int *pci_device;

        *(volatile int*)0xE1000140  = 0;        //Setting footbridge IO ext.to 0

 	// set up ETHER10 address IO address 0x300
 	*(volatile long*)0xE0900010 = 0x301;
        *(volatile char*)0xE0900004 = 1;	//Ether 10

	*(volatile int*) 0xE0180010 = 0x08000000;//intialize VGA PCI mem map
        *(volatile char*)0xE0180004 = 3;	//VGA

	/* Mod made by San, Nov 02, 1998 - Scan PCI IDSEL AD21 for Symbios chip */
	pci_device= (unsigned int*) 0xE0300000;

	if (*pci_device==0x000D1000)
	{
		putstring("setup_arch(): Symbios SCSI / fast ethernet detected\n");
		*(volatile int*) 0xE0300014 = 0x20010000; /* 64k window for registers */
		*(volatile int*) 0xE0300018 = 0x20011000; /* 64k window for scripts   */
		*(volatile char*)0xE0300004 = 0x6;        /* Bus Master + mem mapped  */
	}
	else
	{
		putstring("setup_arch(): Assuming DEC Tulip fast ethernet\n");
    		// Tulip Memory Address : 0x20000000 (Phy Addr: 0xA0000000)
	        *(volatile long*) 0xE0200014 = 0x20000000;
	        *(volatile short*)0xE020003C = 0x19;	//W/H IRQ is 25
		*(volatile char *)0xE020000C = 0x20;	//PCI_CACHE_LINE_SIZE
		*(volatile char *)0xE020000D = 0x20;	//PCI_LATENCY_TIMER
		*(volatile char *)0xE0200004 = 0x6;     /* Bus Master + mem mapped  */
	}
	
//	*(volatile char*)0xE0300004 = 1;	//codec (unused for now)


        *(volatile char*)0xE050010D = 0x20;	//IDE PCI_LATENCY_TIMER
//	*(volatile char*)0xE0500004 = 1;	//ISA (hardwired to 7)
	*(volatile char*)0xE0500104 = 1;        //enable IDE controller as IO

}

struct param_struct *params = (struct param_struct *) PARAMS_BASE;

void setup_arch(char **cmdline_p,
	unsigned long * memory_start_p, unsigned long * memory_end_p)
{
	unsigned long memory_start, memory_end;
	char c = ' ', *to = command_line, *from;
	int len = 0;

	int	i;
	unsigned char	s_num[32];

	memory_start = (unsigned long)&_end;
        memory_end = 0xc2000000;
	from = default_command_line;

	processor = sa110_processor_functions;

	processor._proc_init ();

	min_pci_bios();

        /* get the board ID and tune up the PCI clock */
        if ((board_id & 0xF000)==0x5000
            || (board_id & 0xF000)==0xF000)
            tune_up_pci_clock();
        
        // init the vga to a simple text mode
        vga_init();

        setup_other_io();

        /*
         params support added by SanM - Nov 24th, 1998
         merged into 2.0.35 kernel by RalphS - Jan 6, 1999
         */
        //ROOT_DEV                = 0x0100;
        ROOT_DEV        = to_kdev_t(params->u1.s.rootdev);
        putstring("\nparams root = 0x");
        puthex((int)params->u1.s.rootdev);
        putstring("\n");

        putstring("\nUsing MAJOR of 0x");
        puthex((int)MAJOR(ROOT_DEV));
        putstring(" and minor of 0x");
        puthex((int)MINOR(ROOT_DEV));
        putstring("\n");

        from            = params->commandline;
        memory_end      = GET_MEMORY_END(params);
        putstring("\nMemory end = ");
        puthex((int)memory_end);

	/* Use default command line if param structure is invalid */
        if (((unsigned long) params->u1.s.pages_in_bank[0]!=8192)&&   
	    ((unsigned long) params->u1.s.pages_in_bank[0]!=4096)&&
	    ((unsigned long) params->u1.s.pages_in_bank[0]!=2048)&&
	    ((unsigned long) params->u1.s.pages_in_bank[0]!=1024)) 
          {
            putstring("Params structure invalid. Using default commandline\n");
            from=default_command_line;
            ROOT_DEV=0; /* Gets filled in by parse_params() in start_kernel() (init/main.c) */
          }

	setup_ramdisk();

	/* Save unparsed command line copy for /proc/cmdline */
	memcpy(saved_command_line, from, COMMAND_LINE_SIZE);
	saved_command_line[COMMAND_LINE_SIZE-1] = '\0';

	for (;;) {
		if (c == ' ' &&
		    from[0] == 'm' &&
		    from[1] == 'e' &&
		    from[2] == 'm' &&
		    from[3] == '=') {
			memory_end = simple_strtoul(from+4, &from, 0);
			if ( *from == 'K' || *from == 'k' ) {
				memory_end = memory_end << 10;
				from++;
			} else if ( *from == 'M' || *from == 'm' ) {
				memory_end = memory_end << 20;
				from++;
			}
			memory_end = memory_end + PAGE_OFFSET;
		}
		c = *from++;
		if (!c)
			break;
		if (COMMAND_LINE_SIZE <= ++len)
			break;
		*to++ = c;
	}

	*to = '\0';
	*cmdline_p = command_line;
	*memory_start_p = memory_start;
	*memory_end_p = memory_end;
	strcpy (system_utsname.machine, "armv4l");

	init_task.mm->start_code = TASK_SIZE;
	init_task.mm->end_code	 = TASK_SIZE + (unsigned long) &_etext;
	init_task.mm->end_data	 = TASK_SIZE + (unsigned long) &_edata;
	init_task.mm->brk	 = TASK_SIZE + (unsigned long) &_end;

	putstring("\nboard rev:");
	puthex(board_id);
	
        /* set up the ether 10 and read the serial number */
        outb_p(0x48, 0x30E);
        outb_p(0x00, 0x30A);
        outb_p(0x00, 0x30B);
        outb_p(0x00, 0x30F);
        outb_p(0xFF, 0x307);
        outb_p(0x20, 0x30C);
        outb_p(0x02, 0x30D);
        outb_p(0x20, 0x30A);
        outb_p(0x00, 0x30B);
        outb_p(0x00, 0x308);
        outb_p(0x00, 0x309);
        outb_p(0x0A, 0x300);

        for(i = 0; i < 32; i++)
        {
		s_num[i] = inb(0x310);
        }
	
	if ((s_num[0] == 0) && (s_num[1] == 16) && (s_num[2] == 87))
	{
		serial_num = (s_num[4]<<8) + s_num[5];
	}
	putstring("\nserial #:");
	puthex(serial_num);
}


extern STATS_MISALIGNED Tmisalignstat_stats;
int get_cpuinfo(char * buffer)
{
  int len;

  len = sprintf (buffer,  "CPU:\n"
        "Type\t\t: %s\n"
        "Revision\t: %d\n"
        "Manufacturer\t: %s\n"
        "32bit modes\t: %s\n"
        "BogoMips\t: %lu.%02lu\n"
        "Board Rev\t: %04X\n"
        "Board Serial #\t: %08u\n"
        "\nMisaligned Accesses\t: %d (sys: %d) (usr: %d)\n"
        "  LDR/STR accesses\t: %d\n"
        "  LDRH/STRH accesses\t: %d\n"
        "  LDM/STM accesses\t: %d\n"
        "  SWP accesses    \t: %d\n"
        "  Invalid accesses\t: %d\n",
        "sa110",
        (int)arm_id & 15,
        "DEC",
        "yes",
        (loops_per_sec+2500) / 500000,
        ((loops_per_sec+2500) / 5000) % 100,
        board_id, serial_num,
        (Tmisalignstat_stats.UI_user+Tmisalignstat_stats.UI_sys),
        Tmisalignstat_stats.UI_sys,
        Tmisalignstat_stats.UI_user,
        Tmisalignstat_stats.UI_word,
        Tmisalignstat_stats.UI_short,
        Tmisalignstat_stats.UI_multiple,
        Tmisalignstat_stats.UI_swap,
        Tmisalignstat_stats.UI_unsupported);
  return len;
}

#define BASE (char*)0xd0800000

static void w_seq(int reg, int data) {
 	volatile char *p = BASE;
 	p[0x3c4] = reg;
 	p[0x3c5] = data;
}

static void w_crt(int reg, int data) {
 	volatile char* p = BASE;
 	p[0x3d4] = reg;
 	p[0x3d5] = data;
}
static void w_grf(int reg, int data) {
 	volatile char* p = BASE;
 	p[0x3ce] = reg;
 	p[0x3cf] = data;
}
static void w_atr(int reg, int data) {
 	volatile char*p = BASE;
 	char a;
 	a = p[0x3da];
 	p[0x3c0] = reg;
 	a = p[0x3c1];
 	p[0x3c0] = data;
}

static void delay(volatile int i) { while (i--); }

extern unsigned char chars8x16[256][16];
//char str[] = "another test";
 	
static void vga_init(void) {

 	volatile char* p = BASE;	
 	char a;
 	int i;
 	int j;

 	// wake up the device
 	p[0x46e8] = 0x18;
 	p[0x102] = 1;
 	p[0x46e8] = 8;

putstring("...into vga_init");	
delay(1);
 	p[0x3c2] = 3;		// use B0/B1 clock
 	w_crt(0x11, 0x0c);	// unlock CR0-7
 	w_atr(0x11,0);

delay(3);
 	// set the horizontal and vert totals	
 	w_seq(0,1);
 	w_seq(0,0x3);
 	w_seq(1,0x0);	// NOTE!!! asking for 9 bit dot clock hangs the memory
 	w_seq(2,0xf);	
 	w_seq(3,0);
 	w_seq(4,2);

delay(3);
 	a = p[0x3cc];	
 	p[0x3c2] = 0x67;

 	w_crt(0,0x5f);
 	w_crt(1,0x4f);
 	w_crt(2,0x50);
 	w_crt(3,0x82);  // enable read of cr10/11
 	w_crt(4,0x55);
 	w_crt(5,0x81);
 	w_crt(6,0xbf);
 	w_crt(7,0x1f);
 	w_crt(8,0x0);
 	w_crt(9,0x4f);
 	w_crt(0xA,0xd);	// cursor enabled
 	w_crt(0xB,0x0e);
 	w_crt(0xC,0x0);
 	w_crt(0xD,0x0);
 	w_crt(0xE,0x0);
 	w_crt(0xF,0x0);
 	w_crt(0x10,0x9c);
 	w_crt(0x11,0x0e);
 	w_crt(0x12,0x8f);
 	w_crt(0x13,0x28);
 	w_crt(0x14,0x1f);
 	w_crt(0x15,0x96);
 	w_crt(0x16,0xb9);
 	w_crt(0x17,0xa3);
 	w_crt(0x18,0xff);

 	w_atr(0,0);
 	w_atr(1,1);
 	w_atr(2,2);
 	w_atr(3,3);
 	w_atr(4,4);
 	w_atr(5,5);
 	w_atr(6,6);
 	w_atr(7,7);
 	w_atr(8,0x8);
 	w_atr(9,0x9);
 	w_atr(10,0xa);
 	w_atr(11,0xb);
 	w_atr(12,0xc);
 	w_atr(13,0xd);
 	w_atr(14,0xe);
 	w_atr(15,0xf);
 	w_atr(0x10,0xc);
 	w_atr(0x11,0);
 	w_atr(0x12,0xF);
 	w_atr(0x13,8);
 	w_atr(0x14,0);

 	w_grf(0,0);
 	w_grf(1,0);
 	w_grf(2,0);
 	w_grf(3,0);
 	w_grf(4,0);
 	w_grf(5,0x10);
 	w_grf(6,0x2);
 	w_grf(7,0x0);
 	w_grf(8,0xFF);

delay(3);
 	// init the palette
 	p[0x3c8] = 0;
 	for (i=0; i<255; i++) {
 		// assert the R, G and B
 		p[0x3c9] = 64*i+3*i+(i>>1);
 		p[0x3c9] = 75*i+i;
 		p[0x3c9] = 86*i+2*i+(i>>2);
 		delay(25);
 	}
 	p[0x3c0] = 0x20;
 	p[0x3c6] = 0xFF;
	// and put the palette in low-brightness mode
	// patb 15/10/98
	p[0x3ce] = 0x56;
	i = p[0x3cf];
	p[0x3cf] = i | 4; // unlock 3c6
	p[0x3c6] = 0x04; // low brightness, palette enabled
	p[0x3cf] = i;    // lock it up again

delay(3);

 	// IGS specific extensions
 	w_grf(0x10,0x10);	// ext start addr
 	w_grf(0x11,0x0);	// ext crt vert overfl
 	w_grf(0x12,0x0);	// ext crt irq
 	w_grf(0x13,0x0);	// ext crt test
 	w_grf(0x14,0x51);	// ext num of fetch
 	w_grf(0x15,0x0);
 	w_grf(0x30,0x1);	// ext bus control: 2MB linear ram
 	w_grf(0x31,0x0);	// ext seg write pointer
 	w_grf(0x32,0x0);
 	w_grf(0x33,0x1);	// ext biu misc control; enable linear ram

 	w_grf(0x50,0x0);	// sprite params
 	w_grf(0x51,0x0);
 	w_grf(0x52,0x0);
 	w_grf(0x53,0x0);
 	w_grf(0x54,0x0);
 	w_grf(0x55,0x0);
 	w_grf(0x56,0x0);	
 	w_grf(0x57,0x0);	// ext attrib: standard mode
 	w_grf(0x58,0x0);	// overscans
 	w_grf(0x59,0x0);
 	w_grf(0x5a,0x0);
//// 	w_grf(0x70,0xb);	// ext mem: multiple cas, 7 mclk
//// 	w_grf(0x71,0x10);	//63);	// ext mem: serial access, 2MB x 16
//// 	w_grf(0x72,0x05);	//2);	// ext mem: 32 bit bus, 4M
//	w_grf(0x73,0x30);	
//	w_grf(0x74,0x1b);	// ext mem: fifo depth
//	w_grf(0x75,0x1e);
 	w_grf(0x76,0x0);
 	w_grf(0x77,0x0);	// for low res mode! NOTE: the bios may have set this to non-zero
 	w_grf(0x7a,0xc8);
 	w_grf(0x90,0);		// ext graphics: no bypass gc; NOTE: the bios may have set this to non-zero
 	w_grf(0xBF,0);		// ext graphics: point to real b0..b3 regs (wms)
 	w_grf(0xb0,0xd2);	// pixel clock
 	w_grf(0xb1,0xce);
// 	w_grf(0xb2,0xdb);
// 	w_grf(0xb3,0x54);
 	w_grf(0xb9,0x80);	// latch the freqs into the chip
 	w_grf(0xb9,0);


 	// load font
 	w_seq(0, 1);
         w_seq(2, 4);           /* Enable plane 2 */
 	w_seq(4, 6);		// go to linear mode
 	w_seq(0, 3);
 	w_grf(4, 2);
 	w_grf(5, 0);
 	w_grf(6, 0);
 	p = (volatile unsigned char *) 0xd0000000;
         for (j=0; j<256; j++) {
 		for (i=0; i<16; i++) {
 		        *p++ = chars8x16[j][i];
 		}
 		p+=16;	// only first 32 bytes used
 	}
 	
 	w_seq(0, 1);
 	w_seq(2, 3);	// turn on planes 0 & 1
 	w_seq(4, 3);
 	w_seq(0, 3);
 	w_grf(4, 0);
 	w_grf(5, 0x10);
 	w_grf(6, 2);

 	// lock it up
 	w_crt(0x11, 0x8e);

//wms - cut out the test... Replace it with clear-screen.
 	p = (volatile char*) 0xd0000000;
	while (p < (char *)(0xD0000000+25*80*2))
	{
		*(unsigned int*)p = 0x07200720;	//white spaces on black back
		p += 4;
 	}
}


unsigned char chars8x16[256][16] = {
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 0 */
{0x00,0x00,0x7E,0x81,0xA5,0x81,0x81,0xBD,
0x99,0x81,0x81,0x7E,0x00,0x00,0x00,0x00}, /* 1 */
{0x00,0x00,0x7E,0xFF,0xDB,0xFF,0xFF,0xC3,
0xE7,0xFF,0xFF,0x7E,0x00,0x00,0x00,0x00}, /* 2 */
{0x00,0x00,0x00,0x00,0x6C,0xFE,0xFE,0xFE,
0xFE,0x7C,0x38,0x10,0x00,0x00,0x00,0x00}, /* 3 */
{0x00,0x00,0x00,0x00,0x10,0x38,0x7C,0xFE,
0x7C,0x38,0x10,0x00,0x00,0x00,0x00,0x00}, /* 4 */
{0x00,0x00,0x00,0x18,0x3C,0x3C,0xE7,0xE7,
0xE7,0x99,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 5 */
{0x00,0x00,0x00,0x18,0x3C,0x7E,0xFF,0xFF,
0x7E,0x18,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 6 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x18,0x3C,
0x3C,0x18,0x00,0x00,0x00,0x00,0x00,0x00}, /* 7 */
{0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xE7,0xC3,
0xC3,0xE7,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF}, /* 8 */
{0x00,0x00,0x00,0x00,0x00,0x3C,0x66,0x42,
0x42,0x66,0x3C,0x00,0x00,0x00,0x00,0x00}, /* 9 */
{0xFF,0xFF,0xFF,0xFF,0xFF,0xC3,0x99,0xBD,
0xBD,0x99,0xC3,0xFF,0xFF,0xFF,0xFF,0xFF}, /* 10 */
{0x00,0x00,0x1E,0x0E,0x1A,0x32,0x78,0xCC,
0xCC,0xCC,0xCC,0x78,0x00,0x00,0x00,0x00}, /* 11 */
{0x00,0x00,0x3C,0x66,0x66,0x66,0x66,0x3C,
0x18,0x7E,0x18,0x18,0x00,0x00,0x00,0x00}, /* 12 */
{0x00,0x00,0x3F,0x33,0x3F,0x30,0x30,0x30,
0x30,0x70,0xF0,0xE0,0x00,0x00,0x00,0x00}, /* 13 */
{0x00,0x00,0x7F,0x63,0x7F,0x63,0x63,0x63,
0x63,0x67,0xE7,0xE6,0xC0,0x00,0x00,0x00}, /* 14 */
{0x00,0x00,0x00,0x18,0x18,0xDB,0x3C,0xE7,
0x3C,0xDB,0x18,0x18,0x00,0x00,0x00,0x00}, /* 15 */
{0x00,0x80,0xC0,0xE0,0xF0,0xF8,0xFE,0xF8,
0xF0,0xE0,0xC0,0x80,0x00,0x00,0x00,0x00}, /* 16 */
{0x00,0x02,0x06,0x0E,0x1E,0x3E,0xFE,0x3E,
0x1E,0x0E,0x06,0x02,0x00,0x00,0x00,0x00}, /* 17 */
{0x00,0x00,0x18,0x3C,0x7E,0x18,0x18,0x18,
0x18,0x7E,0x3C,0x18,0x00,0x00,0x00,0x00}, /* 18 */
{0x00,0x00,0x66,0x66,0x66,0x66,0x66,0x66,
0x66,0x00,0x66,0x66,0x00,0x00,0x00,0x00}, /* 19 */
{0x00,0x00,0x7F,0xDB,0xDB,0xDB,0x7B,0x1B,
0x1B,0x1B,0x1B,0x1B,0x00,0x00,0x00,0x00}, /* 20 */
{0x00,0x7C,0xC6,0x60,0x38,0x6C,0xC6,0xC6,
0x6C,0x38,0x0C,0xC6,0x7C,0x00,0x00,0x00}, /* 21 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0xFE,0xFE,0xFE,0xFE,0x00,0x00,0x00,0x00}, /* 22 */
{0x00,0x00,0x18,0x3C,0x7E,0x18,0x18,0x18,
0x18,0x7E,0x3C,0x18,0x7E,0x00,0x00,0x00}, /* 23 */
{0x00,0x00,0x18,0x3C,0x7E,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x00,0x00,0x00,0x00}, /* 24 */
{0x00,0x00,0x18,0x18,0x18,0x18,0x18,0x18,
0x18,0x7E,0x3C,0x18,0x00,0x00,0x00,0x00}, /* 25 */
{0x00,0x00,0x00,0x00,0x00,0x18,0x0C,0xFE,
0x0C,0x18,0x00,0x00,0x00,0x00,0x00,0x00}, /* 26 */
{0x00,0x00,0x00,0x00,0x00,0x30,0x60,0xFE,
0x60,0x30,0x00,0x00,0x00,0x00,0x00,0x00}, /* 27 */
{0x00,0x00,0x00,0x00,0x00,0xC0,0xC0,0xC0,
0xC0,0xFE,0x00,0x00,0x00,0x00,0x00,0x00}, /* 28 */
{0x00,0x00,0x00,0x00,0x00,0x28,0x6C,0xFE,
0x6C,0x28,0x00,0x00,0x00,0x00,0x00,0x00}, /* 29 */
{0x00,0x00,0x00,0x00,0x10,0x38,0x38,0x7C,
0x7C,0xFE,0xFE,0x00,0x00,0x00,0x00,0x00}, /* 30 */
{0x00,0x00,0x00,0x00,0xFE,0xFE,0x7C,0x7C,
0x38,0x38,0x10,0x00,0x00,0x00,0x00,0x00}, /* 31 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 32 */
{0x00,0x00,0x18,0x3C,0x3C,0x3C,0x18,0x18,
0x18,0x00,0x18,0x18,0x00,0x00,0x00,0x00}, /* 33 */
{0x00,0x66,0x66,0x66,0x24,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 34 */
{0x00,0x00,0x00,0x6C,0x6C,0xFE,0x6C,0x6C,
0x6C,0xFE,0x6C,0x6C,0x00,0x00,0x00,0x00}, /* 35 */
{0x18,0x18,0x7C,0xC6,0xC2,0xC0,0x7C,0x06,
0x86,0xC6,0x7C,0x18,0x18,0x00,0x00,0x00}, /* 36 */
{0x00,0x00,0x00,0x00,0xC2,0xC6,0x0C,0x18,
0x30,0x60,0xC6,0x86,0x00,0x00,0x00,0x00}, /* 37 */
{0x00,0x00,0x38,0x6C,0x6C,0x38,0x76,0xDC,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 38 */
{0x00,0x30,0x30,0x30,0x60,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 39 */
{0x00,0x00,0x0C,0x18,0x30,0x30,0x30,0x30,
0x30,0x30,0x18,0x0C,0x00,0x00,0x00,0x00}, /* 40 */
{0x00,0x00,0x30,0x18,0x0C,0x0C,0x0C,0x0C,
0x0C,0x0C,0x18,0x30,0x00,0x00,0x00,0x00}, /* 41 */
{0x00,0x00,0x00,0x00,0x00,0x66,0x3C,0xFF,
0x3C,0x66,0x00,0x00,0x00,0x00,0x00,0x00}, /* 42 */
{0x00,0x00,0x00,0x00,0x00,0x18,0x18,0x7E,
0x18,0x18,0x00,0x00,0x00,0x00,0x00,0x00}, /* 43 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x18,0x18,0x18,0x30,0x00,0x00,0x00}, /* 44 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFE,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 45 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x18,0x18,0x00,0x00,0x00,0x00}, /* 46 */
{0x00,0x00,0x00,0x00,0x02,0x06,0x0C,0x18,
0x30,0x60,0xC0,0x80,0x00,0x00,0x00,0x00}, /* 47 */
{0x00,0x00,0x3C,0x66,0xC3,0xC3,0xDB,0xDB,
0xC3,0xC3,0x66,0x3C,0x00,0x00,0x00,0x00}, /* 48 */
{0x00,0x00,0x18,0x38,0x78,0x18,0x18,0x18,
0x18,0x18,0x18,0x7E,0x00,0x00,0x00,0x00}, /* 49 */
{0x00,0x00,0x7C,0xC6,0x06,0x0C,0x18,0x30,
0x60,0xC0,0xC6,0xFE,0x00,0x00,0x00,0x00}, /* 50 */
{0x00,0x00,0x7C,0xC6,0x06,0x06,0x3C,0x06,
0x06,0x06,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 51 */
{0x00,0x00,0x0C,0x1C,0x3C,0x6C,0xCC,0xFE,
0x0C,0x0C,0x0C,0x1E,0x00,0x00,0x00,0x00}, /* 52 */
{0x00,0x00,0xFE,0xC0,0xC0,0xC0,0xFC,0x06,
0x06,0x06,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 53 */
{0x00,0x00,0x38,0x60,0xC0,0xC0,0xFC,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 54 */
{0x00,0x00,0xFE,0xC6,0x06,0x06,0x0C,0x18,
0x30,0x30,0x30,0x30,0x00,0x00,0x00,0x00}, /* 55 */
{0x00,0x00,0x7C,0xC6,0xC6,0xC6,0x7C,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 56 */
{0x00,0x00,0x7C,0xC6,0xC6,0xC6,0x7E,0x06,
0x06,0x06,0x0C,0x78,0x00,0x00,0x00,0x00}, /* 57 */
{0x00,0x00,0x00,0x00,0x18,0x18,0x00,0x00,
0x00,0x18,0x18,0x00,0x00,0x00,0x00,0x00}, /* 58 */
{0x00,0x00,0x00,0x00,0x18,0x18,0x00,0x00,
0x00,0x18,0x18,0x30,0x00,0x00,0x00,0x00}, /* 59 */
{0x00,0x00,0x00,0x06,0x0C,0x18,0x30,0x60,
0x30,0x18,0x0C,0x06,0x00,0x00,0x00,0x00}, /* 60 */
{0x00,0x00,0x00,0x00,0x00,0x00,0xFE,0x00,
0x00,0xFE,0x00,0x00,0x00,0x00,0x00,0x00}, /* 61 */
{0x00,0x00,0x00,0x60,0x30,0x18,0x0C,0x06,
0x0C,0x18,0x30,0x60,0x00,0x00,0x00,0x00}, /* 62 */
{0x00,0x00,0x7C,0xC6,0xC6,0x0C,0x18,0x18,
0x18,0x00,0x18,0x18,0x00,0x00,0x00,0x00}, /* 63 */
{0x00,0x00,0x00,0x7C,0xC6,0xC6,0xDE,0xDE,
0xDE,0xDC,0xC0,0x7C,0x00,0x00,0x00,0x00}, /* 64 */
{0x00,0x00,0x10,0x38,0x6C,0xC6,0xC6,0xFE,
0xC6,0xC6,0xC6,0xC6,0x00,0x00,0x00,0x00}, /* 65 */
{0x00,0x00,0xFC,0x66,0x66,0x66,0x7C,0x66,
0x66,0x66,0x66,0xFC,0x00,0x00,0x00,0x00}, /* 66 */
{0x00,0x00,0x3C,0x66,0xC2,0xC0,0xC0,0xC0,
0xC0,0xC2,0x66,0x3C,0x00,0x00,0x00,0x00}, /* 67 */
{0x00,0x00,0xF8,0x6C,0x66,0x66,0x66,0x66,
0x66,0x66,0x6C,0xF8,0x00,0x00,0x00,0x00}, /* 68 */
{0x00,0x00,0xFE,0x66,0x62,0x68,0x78,0x68,
0x60,0x62,0x66,0xFE,0x00,0x00,0x00,0x00}, /* 69 */
{0x00,0x00,0xFE,0x66,0x62,0x68,0x78,0x68,
0x60,0x60,0x60,0xF0,0x00,0x00,0x00,0x00}, /* 70 */
{0x00,0x00,0x3C,0x66,0xC2,0xC0,0xC0,0xDE,
0xC6,0xC6,0x66,0x3A,0x00,0x00,0x00,0x00}, /* 71 */
{0x00,0x00,0xC6,0xC6,0xC6,0xC6,0xFE,0xC6,
0xC6,0xC6,0xC6,0xC6,0x00,0x00,0x00,0x00}, /* 72 */
{0x00,0x00,0x3C,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 73 */
{0x00,0x00,0x1E,0x0C,0x0C,0x0C,0x0C,0x0C,
0xCC,0xCC,0xCC,0x78,0x00,0x00,0x00,0x00}, /* 74 */
{0x00,0x00,0xE6,0x66,0x6C,0x6C,0x78,0x78,
0x6C,0x66,0x66,0xE6,0x00,0x00,0x00,0x00}, /* 75 */
{0x00,0x00,0xF0,0x60,0x60,0x60,0x60,0x60,
0x60,0x62,0x66,0xFE,0x00,0x00,0x00,0x00}, /* 76 */
{0x00,0x00,0xC6,0xEE,0xFE,0xFE,0xD6,0xC6,
0xC6,0xC6,0xC6,0xC6,0x00,0x00,0x00,0x00}, /* 77 */
{0x00,0x00,0xC6,0xE6,0xF6,0xFE,0xDE,0xCE,
0xC6,0xC6,0xC6,0xC6,0x00,0x00,0x00,0x00}, /* 78 */
{0x00,0x00,0x7C,0xC6,0xC6,0xC6,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 79 */
{0x00,0x00,0xFC,0x66,0x66,0x66,0x7C,0x60,
0x60,0x60,0x60,0xF0,0x00,0x00,0x00,0x00}, /* 80 */
{0x00,0x00,0x7C,0xC6,0xC6,0xC6,0xC6,0xC6,
0xC6,0xD6,0xDE,0x7C,0x0C,0x0E,0x00,0x00}, /* 81 */
{0x00,0x00,0xFC,0x66,0x66,0x66,0x7C,0x6C,
0x66,0x66,0x66,0xE6,0x00,0x00,0x00,0x00}, /* 82 */
{0x00,0x00,0x7C,0xC6,0xC6,0x60,0x38,0x0C,
0x06,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 83 */
{0x00,0x00,0x7E,0x7E,0x5A,0x18,0x18,0x18,
0x18,0x18,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 84 */
{0x00,0x00,0xC6,0xC6,0xC6,0xC6,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 85 */
{0x00,0x00,0xC6,0xC6,0xC6,0xC6,0xC6,0xC6,
0xC6,0x6C,0x38,0x10,0x00,0x00,0x00,0x00}, /* 86 */
{0x00,0x00,0xC6,0xC6,0xC6,0xC6,0xC6,0xD6,
0xD6,0xFE,0x6C,0x6C,0x00,0x00,0x00,0x00}, /* 87 */
{0x00,0x00,0xC6,0xC6,0x6C,0x6C,0x38,0x38,
0x6C,0x6C,0xC6,0xC6,0x00,0x00,0x00,0x00}, /* 88 */
{0x00,0x00,0x66,0x66,0x66,0x66,0x3C,0x18,
0x18,0x18,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 89 */
{0x00,0x00,0xFE,0xC6,0x86,0x0C,0x18,0x30,
0x60,0xC2,0xC6,0xFE,0x00,0x00,0x00,0x00}, /* 90 */
{0x00,0x00,0x3C,0x30,0x30,0x30,0x30,0x30,
0x30,0x30,0x30,0x3C,0x00,0x00,0x00,0x00}, /* 91 */
{0x00,0x00,0x00,0x80,0xC0,0xE0,0x70,0x38,
0x1C,0x0E,0x06,0x02,0x00,0x00,0x00,0x00}, /* 92 */
{0x00,0x00,0x3C,0x0C,0x0C,0x0C,0x0C,0x0C,
0x0C,0x0C,0x0C,0x3C,0x00,0x00,0x00,0x00}, /* 93 */
{0x10,0x38,0x6C,0xC6,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 94 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0xFF,0x00,0x00}, /* 95 */
{0x30,0x30,0x18,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 96 */
{0x00,0x00,0x00,0x00,0x00,0x78,0x0C,0x7C,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 97 */
{0x00,0x00,0xE0,0x60,0x60,0x78,0x6C,0x66,
0x66,0x66,0x66,0xDC,0x00,0x00,0x00,0x00}, /* 98 */
{0x00,0x00,0x00,0x00,0x00,0x7C,0xC6,0xC0,
0xC0,0xC0,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 99 */
{0x00,0x00,0x1C,0x0C,0x0C,0x3C,0x6C,0xCC,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 100 */
{0x00,0x00,0x00,0x00,0x00,0x7C,0xC6,0xFE,
0xC0,0xC0,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 101 */
{0x00,0x00,0x38,0x6C,0x64,0x60,0xF0,0x60,
0x60,0x60,0x60,0xF0,0x00,0x00,0x00,0x00}, /* 102 */
{0x00,0x00,0x00,0x00,0x00,0x76,0xCC,0xCC,
0xCC,0xCC,0xCC,0x7C,0x0C,0xCC,0x78,0x00}, /* 103 */
{0x00,0x00,0xE0,0x60,0x60,0x6C,0x76,0x66,
0x66,0x66,0x66,0xE6,0x00,0x00,0x00,0x00}, /* 104 */
{0x00,0x00,0x18,0x18,0x00,0x38,0x18,0x18,
0x18,0x18,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 105 */
{0x00,0x00,0x06,0x06,0x00,0x0E,0x06,0x06,
0x06,0x06,0x06,0x06,0x66,0x66,0x3C,0x00}, /* 106 */
{0x00,0x00,0xE0,0x60,0x60,0x66,0x6C,0x78,
0x78,0x6C,0x66,0xE6,0x00,0x00,0x00,0x00}, /* 107 */
{0x00,0x00,0x38,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 108 */
{0x00,0x00,0x00,0x00,0x00,0xEC,0xFE,0xD6,
0xD6,0xD6,0xD6,0xD6,0x00,0x00,0x00,0x00}, /* 109 */
{0x00,0x00,0x00,0x00,0x00,0xDC,0x66,0x66,
0x66,0x66,0x66,0x66,0x00,0x00,0x00,0x00}, /* 110 */
{0x00,0x00,0x00,0x00,0x00,0x7C,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 111 */
{0x00,0x00,0x00,0x00,0x00,0xDC,0x66,0x66,
0x66,0x66,0x66,0x7C,0x60,0x60,0xF0,0x00}, /* 112 */
{0x00,0x00,0x00,0x00,0x00,0x76,0xCC,0xCC,
0xCC,0xCC,0xCC,0x7C,0x0C,0x0C,0x1E,0x00}, /* 113 */
{0x00,0x00,0x00,0x00,0x00,0xDC,0x76,0x62,
0x60,0x60,0x60,0xF0,0x00,0x00,0x00,0x00}, /* 114 */
{0x00,0x00,0x00,0x00,0x00,0x7C,0xC6,0x60,
0x38,0x0C,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 115 */
{0x00,0x00,0x10,0x30,0x30,0xFC,0x30,0x30,
0x30,0x30,0x36,0x1C,0x00,0x00,0x00,0x00}, /* 116 */
{0x00,0x00,0x00,0x00,0x00,0xCC,0xCC,0xCC,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 117 */
{0x00,0x00,0x00,0x00,0x00,0x66,0x66,0x66,
0x66,0x66,0x3C,0x18,0x00,0x00,0x00,0x00}, /* 118 */
{0x00,0x00,0x00,0x00,0x00,0xC6,0xC6,0xC6,
0xD6,0xD6,0xFE,0x6C,0x00,0x00,0x00,0x00}, /* 119 */
{0x00,0x00,0x00,0x00,0x00,0xC6,0x6C,0x38,
0x38,0x38,0x6C,0xC6,0x00,0x00,0x00,0x00}, /* 120 */
{0x00,0x00,0x00,0x00,0x00,0xC6,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7E,0x06,0x0C,0xF8,0x00}, /* 121 */
{0x00,0x00,0x00,0x00,0x00,0xFE,0xCC,0x18,
0x30,0x60,0xC6,0xFE,0x00,0x00,0x00,0x00}, /* 122 */
{0x00,0x00,0x0E,0x18,0x18,0x18,0x70,0x18,
0x18,0x18,0x18,0x0E,0x00,0x00,0x00,0x00}, /* 123 */
{0x00,0x00,0x18,0x18,0x18,0x18,0x00,0x18,
0x18,0x18,0x18,0x18,0x00,0x00,0x00,0x00}, /* 124 */
{0x00,0x00,0x70,0x18,0x18,0x18,0x0E,0x18,
0x18,0x18,0x18,0x70,0x00,0x00,0x00,0x00}, /* 125 */
{0x00,0x00,0x76,0xDC,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 126 */
{0x00,0x00,0x00,0x00,0x10,0x38,0x6C,0xC6,
0xC6,0xC6,0xFE,0x00,0x00,0x00,0x00,0x00}, /* 127 */
{0x00,0x00,0x3C,0x66,0xC2,0xC0,0xC0,0xC0,
0xC2,0x66,0x3C,0x0C,0x06,0x7C,0x00,0x00}, /* 128 */
{0x00,0x00,0xCC,0x00,0x00,0xCC,0xCC,0xCC,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 129 */
{0x00,0x0C,0x18,0x30,0x00,0x7C,0xC6,0xFE,
0xC0,0xC0,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 130 */
{0x00,0x10,0x38,0x6C,0x00,0x78,0x0C,0x7C,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 131 */
{0x00,0x00,0xCC,0x00,0x00,0x78,0x0C,0x7C,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 132 */
{0x00,0x60,0x30,0x18,0x00,0x78,0x0C,0x7C,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 133 */
{0x00,0x38,0x6C,0x38,0x00,0x78,0x0C,0x7C,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 134 */
{0x00,0x00,0x00,0x00,0x3C,0x66,0x60,0x60,
0x66,0x3C,0x0C,0x06,0x3C,0x00,0x00,0x00}, /* 135 */
{0x00,0x10,0x38,0x6C,0x00,0x7C,0xC6,0xFE,
0xC0,0xC0,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 136 */
{0x00,0x00,0xC6,0x00,0x00,0x7C,0xC6,0xFE,
0xC0,0xC0,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 137 */
{0x00,0x60,0x30,0x18,0x00,0x7C,0xC6,0xFE,
0xC0,0xC0,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 138 */
{0x00,0x00,0x66,0x00,0x00,0x38,0x18,0x18,
0x18,0x18,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 139 */
{0x00,0x18,0x3C,0x66,0x00,0x38,0x18,0x18,
0x18,0x18,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 140 */
{0x00,0x60,0x30,0x18,0x00,0x38,0x18,0x18,
0x18,0x18,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 141 */
{0x00,0xC6,0x00,0x10,0x38,0x6C,0xC6,0xC6,
0xFE,0xC6,0xC6,0xC6,0x00,0x00,0x00,0x00}, /* 142 */
{0x38,0x6C,0x38,0x00,0x38,0x6C,0xC6,0xC6,
0xFE,0xC6,0xC6,0xC6,0x00,0x00,0x00,0x00}, /* 143 */
{0x18,0x30,0x60,0x00,0xFE,0x66,0x60,0x7C,
0x60,0x60,0x66,0xFE,0x00,0x00,0x00,0x00}, /* 144 */
{0x00,0x00,0x00,0x00,0x00,0xCC,0x76,0x36,
0x7E,0xD8,0xD8,0x6E,0x00,0x00,0x00,0x00}, /* 145 */
{0x00,0x00,0x3E,0x6C,0xCC,0xCC,0xFE,0xCC,
0xCC,0xCC,0xCC,0xCE,0x00,0x00,0x00,0x00}, /* 146 */
{0x00,0x10,0x38,0x6C,0x00,0x7C,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 147 */
{0x00,0x00,0xC6,0x00,0x00,0x7C,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 148 */
{0x00,0x60,0x30,0x18,0x00,0x7C,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 149 */
{0x00,0x30,0x78,0xCC,0x00,0xCC,0xCC,0xCC,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 150 */
{0x00,0x60,0x30,0x18,0x00,0xCC,0xCC,0xCC,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 151 */
{0x00,0x00,0xC6,0x00,0x00,0xC6,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7E,0x06,0x0C,0x78,0x00}, /* 152 */
{0x00,0xC6,0x00,0x7C,0xC6,0xC6,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 153 */
{0x00,0xC6,0x00,0xC6,0xC6,0xC6,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 154 */
{0x00,0x18,0x18,0x3C,0x66,0x60,0x60,0x60,
0x66,0x3C,0x18,0x18,0x00,0x00,0x00,0x00}, /* 155 */
{0x00,0x38,0x6C,0x64,0x60,0xF0,0x60,0x60,
0x60,0x60,0xE6,0xFC,0x00,0x00,0x00,0x00}, /* 156 */
{0x00,0x00,0x66,0x66,0x3C,0x18,0x7E,0x18,
0x7E,0x18,0x18,0x18,0x00,0x00,0x00,0x00}, /* 157 */
{0x00,0xF8,0xCC,0xCC,0xF8,0xC4,0xCC,0xDE,
0xCC,0xCC,0xCC,0xC6,0x00,0x00,0x00,0x00}, /* 158 */
{0x00,0x0E,0x1B,0x18,0x18,0x18,0x7E,0x18,
0x18,0x18,0x18,0x18,0xD8,0x70,0x00,0x00}, /* 159 */
{0x00,0x18,0x30,0x60,0x00,0x78,0x0C,0x7C,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 160 */
{0x00,0x0C,0x18,0x30,0x00,0x38,0x18,0x18,
0x18,0x18,0x18,0x3C,0x00,0x00,0x00,0x00}, /* 161 */
{0x00,0x18,0x30,0x60,0x00,0x7C,0xC6,0xC6,
0xC6,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 162 */
{0x00,0x18,0x30,0x60,0x00,0xCC,0xCC,0xCC,
0xCC,0xCC,0xCC,0x76,0x00,0x00,0x00,0x00}, /* 163 */
{0x00,0x00,0x76,0xDC,0x00,0xDC,0x66,0x66,
0x66,0x66,0x66,0x66,0x00,0x00,0x00,0x00}, /* 164 */
{0x76,0xDC,0x00,0xC6,0xE6,0xF6,0xFE,0xDE,
0xCE,0xC6,0xC6,0xC6,0x00,0x00,0x00,0x00}, /* 165 */
{0x00,0x3C,0x6C,0x6C,0x3E,0x00,0x7E,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 166 */
{0x00,0x38,0x6C,0x6C,0x38,0x00,0x7C,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 167 */
{0x00,0x00,0x30,0x30,0x00,0x30,0x30,0x60,
0xC0,0xC6,0xC6,0x7C,0x00,0x00,0x00,0x00}, /* 168 */
{0x00,0x00,0x00,0x00,0x00,0x00,0xFE,0xC0,
0xC0,0xC0,0xC0,0x00,0x00,0x00,0x00,0x00}, /* 169 */
{0x00,0x00,0x00,0x00,0x00,0x00,0xFE,0x06,
0x06,0x06,0x06,0x00,0x00,0x00,0x00,0x00}, /* 170 */
{0x00,0xC0,0xC0,0xC2,0xC6,0xCC,0x18,0x30,
0x60,0xCE,0x93,0x06,0x0C,0x1F,0x00,0x00}, /* 171 */
{0x00,0xC0,0xC0,0xC2,0xC6,0xCC,0x18,0x30,
0x66,0xCE,0x9A,0x3F,0x06,0x0F,0x00,0x00}, /* 172 */
{0x00,0x00,0x18,0x18,0x00,0x18,0x18,0x18,
0x3C,0x3C,0x3C,0x18,0x00,0x00,0x00,0x00}, /* 173 */
{0x00,0x00,0x00,0x00,0x00,0x33,0x66,0xCC,
0x66,0x33,0x00,0x00,0x00,0x00,0x00,0x00}, /* 174 */
{0x00,0x00,0x00,0x00,0x00,0xCC,0x66,0x33,
0x66,0xCC,0x00,0x00,0x00,0x00,0x00,0x00}, /* 175 */
{0x11,0x44,0x11,0x44,0x11,0x44,0x11,0x44,
0x11,0x44,0x11,0x44,0x11,0x44,0x11,0x44}, /* 176 */
{0x55,0xAA,0x55,0xAA,0x55,0xAA,0x55,0xAA,
0x55,0xAA,0x55,0xAA,0x55,0xAA,0x55,0xAA}, /* 177 */
{0xDD,0x77,0xDD,0x77,0xDD,0x77,0xDD,0x77,
0xDD,0x77,0xDD,0x77,0xDD,0x77,0xDD,0x77}, /* 178 */
{0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 179 */
{0x18,0x18,0x18,0x18,0x18,0x18,0x18,0xF8,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 180 */
{0x18,0x18,0x18,0x18,0x18,0xF8,0x18,0xF8,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 181 */
{0x36,0x36,0x36,0x36,0x36,0x36,0x36,0xF6,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 182 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFE,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 183 */
{0x00,0x00,0x00,0x00,0x00,0xF8,0x18,0xF8,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 184 */
{0x36,0x36,0x36,0x36,0x36,0xF6,0x06,0xF6,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 185 */
{0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 186 */
{0x00,0x00,0x00,0x00,0x00,0xFE,0x06,0xF6,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 187 */
{0x36,0x36,0x36,0x36,0x36,0xF6,0x06,0xFE,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 188 */
{0x36,0x36,0x36,0x36,0x36,0x36,0x36,0xFE,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 189 */
{0x18,0x18,0x18,0x18,0x18,0xF8,0x18,0xF8,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 190 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xF8,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 191 */
{0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x1F,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 192 */
{0x18,0x18,0x18,0x18,0x18,0x18,0x18,0xFF,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 193 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 194 */
{0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x1F,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 195 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 196 */
{0x18,0x18,0x18,0x18,0x18,0x18,0x18,0xFF,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 197 */
{0x18,0x18,0x18,0x18,0x18,0x1F,0x18,0x1F,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 198 */
{0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x37,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 199 */
{0x36,0x36,0x36,0x36,0x36,0x37,0x30,0x3F,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 200 */
{0x00,0x00,0x00,0x00,0x00,0x3F,0x30,0x37,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 201 */
{0x36,0x36,0x36,0x36,0x36,0xF7,0x00,0xFF,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 202 */
{0x00,0x00,0x00,0x00,0x00,0xFF,0x00,0xF7,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 203 */
{0x36,0x36,0x36,0x36,0x36,0x37,0x30,0x37,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 204 */
{0x00,0x00,0x00,0x00,0x00,0xFF,0x00,0xFF,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 205 */
{0x36,0x36,0x36,0x36,0x36,0xF7,0x00,0xF7,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 206 */
{0x18,0x18,0x18,0x18,0x18,0xFF,0x00,0xFF,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 207 */
{0x36,0x36,0x36,0x36,0x36,0x36,0x36,0xFF,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 208 */
{0x00,0x00,0x00,0x00,0x00,0xFF,0x00,0xFF,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 209 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 210 */
{0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x3F,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 211 */
{0x18,0x18,0x18,0x18,0x18,0x1F,0x18,0x1F,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 212 */
{0x00,0x00,0x00,0x00,0x00,0x1F,0x18,0x1F,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 213 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x3F,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 214 */
{0x36,0x36,0x36,0x36,0x36,0x36,0x36,0xFF,
0x36,0x36,0x36,0x36,0x36,0x36,0x36,0x36}, /* 215 */
{0x18,0x18,0x18,0x18,0x18,0xFF,0x18,0xFF,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 216 */
{0x18,0x18,0x18,0x18,0x18,0x18,0x18,0xF8,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 217 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x1F,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 218 */
{0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,
0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF}, /* 219 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,
0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF}, /* 220 */
{0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,
0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0,0xF0}, /* 221 */
{0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,
0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F,0x0F}, /* 222 */
{0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 223 */
{0x00,0x00,0x00,0x00,0x00,0x76,0xDC,0xD8,
0xD8,0xD8,0xDC,0x76,0x00,0x00,0x00,0x00}, /* 224 */
{0x00,0x00,0x78,0xCC,0xCC,0xCC,0xD8,0xCC,
0xC6,0xC6,0xC6,0xCC,0x00,0x00,0x00,0x00}, /* 225 */
{0x00,0x00,0xFE,0xC6,0xC6,0xC0,0xC0,0xC0,
0xC0,0xC0,0xC0,0xC0,0x00,0x00,0x00,0x00}, /* 226 */
{0x00,0x00,0x00,0x00,0xFE,0x6C,0x6C,0x6C,
0x6C,0x6C,0x6C,0x6C,0x00,0x00,0x00,0x00}, /* 227 */
{0x00,0x00,0x00,0xFE,0xC6,0x60,0x30,0x18,
0x30,0x60,0xC6,0xFE,0x00,0x00,0x00,0x00}, /* 228 */
{0x00,0x00,0x00,0x00,0x00,0x7E,0xD8,0xD8,
0xD8,0xD8,0xD8,0x70,0x00,0x00,0x00,0x00}, /* 229 */
{0x00,0x00,0x00,0x00,0x66,0x66,0x66,0x66,
0x66,0x7C,0x60,0x60,0xC0,0x00,0x00,0x00}, /* 230 */
{0x00,0x00,0x00,0x00,0x76,0xDC,0x18,0x18,
0x18,0x18,0x18,0x18,0x00,0x00,0x00,0x00}, /* 231 */
{0x00,0x00,0x00,0x7E,0x18,0x3C,0x66,0x66,
0x66,0x3C,0x18,0x7E,0x00,0x00,0x00,0x00}, /* 232 */
{0x00,0x00,0x00,0x38,0x6C,0xC6,0xC6,0xFE,
0xC6,0xC6,0x6C,0x38,0x00,0x00,0x00,0x00}, /* 233 */
{0x00,0x00,0x38,0x6C,0xC6,0xC6,0xC6,0x6C,
0x6C,0x6C,0x6C,0xEE,0x00,0x00,0x00,0x00}, /* 234 */
{0x00,0x00,0x1E,0x30,0x18,0x0C,0x3E,0x66,
0x66,0x66,0x66,0x3C,0x00,0x00,0x00,0x00}, /* 235 */
{0x00,0x00,0x00,0x00,0x00,0x7E,0xDB,0xDB,
0xDB,0x7E,0x00,0x00,0x00,0x00,0x00,0x00}, /* 236 */
{0x00,0x00,0x00,0x03,0x06,0x7E,0xCF,0xDB,
0xF3,0x7E,0x60,0xC0,0x00,0x00,0x00,0x00}, /* 237 */
{0x00,0x00,0x1C,0x30,0x60,0x60,0x7C,0x60,
0x60,0x60,0x30,0x1C,0x00,0x00,0x00,0x00}, /* 238 */
{0x00,0x00,0x00,0x7C,0xC6,0xC6,0xC6,0xC6,
0xC6,0xC6,0xC6,0xC6,0x00,0x00,0x00,0x00}, /* 239 */
{0x00,0x00,0x00,0x00,0xFE,0x00,0x00,0xFE,
0x00,0x00,0xFE,0x00,0x00,0x00,0x00,0x00}, /* 240 */
{0x00,0x00,0x00,0x00,0x18,0x18,0x7E,0x18,
0x18,0x00,0x00,0xFF,0x00,0x00,0x00,0x00}, /* 241 */
{0x00,0x00,0x00,0x30,0x18,0x0C,0x06,0x0C,
0x18,0x30,0x00,0x7E,0x00,0x00,0x00,0x00}, /* 242 */
{0x00,0x00,0x00,0x0C,0x18,0x30,0x60,0x30,
0x18,0x0C,0x00,0x7E,0x00,0x00,0x00,0x00}, /* 243 */
{0x00,0x00,0x0E,0x1B,0x1B,0x18,0x18,0x18,
0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18}, /* 244 */
{0x18,0x18,0x18,0x18,0x18,0x18,0x18,0x18,
0xD8,0xD8,0xD8,0x70,0x00,0x00,0x00,0x00}, /* 245 */
{0x00,0x00,0x00,0x00,0x18,0x18,0x00,0x7E,
0x00,0x18,0x18,0x00,0x00,0x00,0x00,0x00}, /* 246 */
{0x00,0x00,0x00,0x00,0x00,0x76,0xDC,0x00,
0x76,0xDC,0x00,0x00,0x00,0x00,0x00,0x00}, /* 247 */
{0x00,0x38,0x6C,0x6C,0x38,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 248 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x18,
0x18,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 249 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x18,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 250 */
{0x00,0x0F,0x0C,0x0C,0x0C,0x0C,0x0C,0xEC,
0x6C,0x6C,0x3C,0x1C,0x00,0x00,0x00,0x00}, /* 251 */
{0x00,0xD8,0x6C,0x6C,0x6C,0x6C,0x6C,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 252 */
{0x00,0x70,0x98,0x30,0x60,0xC8,0xF8,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}, /* 253 */
{0x00,0x00,0x00,0x00,0x7C,0x7C,0x7C,0x7C,
0x7C,0x7C,0x7C,0x00,0x00,0x00,0x00,0x00}, /* 254 */
{0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}}; /* 255 */

char Read_Byte (void)
{
int lp=2;
char bit=3;
char dest=0;

        for (lp=0; lp<8; lp++)
	{
    	        delay (1000);
		bit=inb(0x203);
//		hex(bit); f(' ');
            	delay(1000);
		bit=inb(0x203);  	//second read
//      	hex(bit); f(' ');
		dest |= (bit << lp);
	}
	return dest;
} //end of read byte



void setup_other_io() 
{
 	int c;

	char si[16];	//serial identifier array	
	int    x,t,CSN,j;
	unsigned char	cTest;


	cShadowX = 0;
	
/*

Woody 26.Feb.98 for Rev.4 H/W

We use GP pins:

GP10 for reset button

GP11 for green LED

GP12 for to disable the fan

//GP13, GP15 and GP23 for upload of codec FPGA (unused)

GP14, GP15 and GP20 for flash write, mute, Phillips 7111 and temprst

GP16 Watchdog (if not used - set for input)

GP17 for red LED

GP21 for E*2 clock

GP22 for temperature sensor

Addressing:
GP10 - RESET (input)	Bit 0, 0x338
GP11 - GRN_PWR_LED	Bit 1
GP12 - FAN-ON		Bit 2
GP13 - DONE (input)	Bit 3
GP14 - IOCLK		Bit 4
GP15 - DATA		Bit 5
GP16 - WD-TIMER		Bit 6
GP17 - RED_PWR_LED	Bit 7

GP20 - IOLOAD		Bit 0, 0x33A
GP21 - E2CLK		Bit 1
GP22 - TEMPCLK		Bit 2
GP23 - CCLK		Bit 3

			Bit 30	of cShadowX = SW flag to display CPU load
			Bit 31	of cShadowX = software LOCK - CPLD in use
*/

 	// open up the SuperIO chip
 	outb(0x87,0x370);
 	outb(0x87,0x370);


 	// set the serial interrupt numbers
 	outb(0x07,0x370);	// device number register
 	outb(0x05,0x371);	// keyboard subsection
 	outb(0x72,0x370);	// interrupt number register for mouse
 	outb(0x05,0x371);	// irq 5

 	outb(0x07,0x370);	// device number
 	outb(0x06,0x371);	// infrared
 	outb(0x70,0x370);	// interrupt number
 	outb(0x06,0x371);	// irq 6
 	

	outb(0x2A,0x370);
	outb(0xC1,0x371);	//enable GP12, GP11, GP10 as I/O, CIRRX, IRRXH

	outb(0x2B,0x370);
	outb(0x6B,0x371);	//enable GP23, GP22, GP21, GP20, GP13 as I/O

 	outb(0x2C,0x370);
        outb(0x55,0x371);	//enable GP17, GP16, GP15, GP14 as I/O
  


 	outb(0x07,0x370);
 	outb(0x07,0x371);	// aux function group 1 (Logical Device 7)

 	outb(0x60,0x370);	// this group is controlled by io port 0x338
 	outb(0x03,0x371);	// GP 11,12,13,14,15,16
 	outb(0x61,0x370);
 	outb(0x38,0x371);

 	outb(0x70,0x370);
 	outb(0x0A,0x371); 	// irq 10 for GP10 (Orange button)

 	outb(0xE0,0x370);	// GP10 control reg
 	outb(0x19,0x371);	// set for debounce & input
  
 	outb(0xE1,0x370);
 	outb(0x00,0x371);	// set GP11 for LED output (green LED)

 	outb(0xE2,0x370);
 	outb(0x00,0x371);	// set GP12 for output (FAN ON)

	outb(0xE3,0x370);	// set GP13 for input (DONE)
	outb(0x01,0x371);

 	outb(0xE4,0x370);
 	outb(0x00,0x371);	// set GP14 for output (IOCLK)

 	outb(0xE5,0x370);
 	outb(0x00,0x371);	// set GP15 for output (DATA)

 	outb(0xE6,0x370);
 	outb(0x01,0x371);	// set GP16 for input (safe), else WD-TIMER

 	outb(0xE7,0x370);
 	outb(0x00,0x371);	// set GP17 for output mode (RED_PWR_LED)

 	outb(0x30,0x370);
 	outb(0x01,0x371);	// turn on section 7 (aux function group 1)

 	outb(0x86,0x338);	// and write out state RedLed/GreenLed/Fan on


	outb(0x07,0x370);
	outb(0x08,0x371);	//select device 8 - Aux II

 	outb(0x60,0x370);	// this group is controlled by io port 0x33A
 	outb(0x03,0x371);
 	outb(0x61,0x370);
 	outb(0x3A,0x371);

 	outb(0xE8,0x370);
 	outb(0x00,0x371);	// set GP20 for output (IOLOAD)

 	outb(0xE9,0x370);
 	outb(0x00,0x371);	// set GP21 for output (E2CLK)

 	outb(0xEA,0x370);
 	outb(0x00,0x371);	// set GP22 for output (TEMPCLK)

 	outb(0xEB,0x370);	// GP23 control reg (CCLK)
 	outb(0x00,0x371);	// enable as output at IO address 0x33A


//clear WatchDog timer regs
 	outb(0xF2,0x370);	// WatchDog Time-Out value
 	outb(0x00,0x371);	// 0 == disabled

 	outb(0xF3,0x370);	// WatchDog reg
 	outb(0x00,0x371);	// reset to default
	
 	outb(0xF4,0x370);	// WatchDog reg
 	outb(0x00,0x371);	// reset if in TIMEOUT'ed state (bit 0)


// T.B.D. set IRDA inputs (touch reg 2A, EC, ED)

 	outb(0x30,0x370);
 	outb(0x01,0x371);	// turn on section 8 (aux function group 2)

	outb(0,0x33A);		// default state



	outb(0x07,0x370);	// select logical device 1 (PPort)
	outb(0x01,0x371); 

	outb(0xF0,0x370);
	outb(0x01,0x371);       // select mode 0x01 == EPP


/*
//these 2 steps are not necessary, as long as pin 44 (PNPCSV) is not connected
 	outb(0x60,0x370);	// this port will appear at io addr 0x378
 	outb(0x03,0x371);
 	outb(0x61,0x370);
 	outb(0x78,0x371);

 	outb(0x30,0x370);
 	outb(0x01,0x371);	// turn on section 1
*/


 	// and close up the EFER gate
 	outb(0xAA,0x370);



 	// turn on the RTC
 	outb(13,0x70);
 	c = inb(0x71);
 	if (!(c & 0x80)) { putstring("\nRTC CMOS batt signal bad\n"); }
 	outb(0x0A,0x70);	// select control reg
 	outb(0x20,0x71);	// make sure the divider is set
 	outb(0x0B,0x70);	// select other control reg
 	c = inb(0x71);		// read it
 	c &= 0xF9; c |= 2;	// 24 hour mode, BCD, update enabled
 	outb(0x0B,0x70);
 	outb(c,0x71);		// turn on BCD counting and 24 clock mode

	
	//we want to write a bit pattern 0110 to Xilinx for DS1620 rst
	//unmute the mono speaker, disable Phillips and disable flash write
	
	set_CPLD(0x06,0x7FFFFFFF);
	
	
// Woody 20.Jan.1998		
// Rockwell sound chip RWA010 PNP stuff comes here....	
	
	outb(0x02,0x279);	//	*c=0x02;
	outb(0x02,0xa79);
	delay(10000);

	//Initiation LFSR Function

	outb(0x00,0x279); outb(0x00,0x279);
	outb(0x6A,0x279); outb(0xb5,0x279); outb(0xDA,0x279); outb(0xED,0x279); 
	outb(0xF6,0x279); outb(0xFB,0x279); outb(0x7D,0x279); outb(0xBE,0x279);
	outb(0xDF,0x279); outb(0x6F,0x279); outb(0x37,0x279); outb(0x1B,0x279);
	outb(0x0D,0x279); outb(0x86,0x279); outb(0xC3,0x279); outb(0x61,0x279);
	outb(0xB0,0x279); outb(0x58,0x279); outb(0x2C,0x279); outb(0x16,0x279); 
	outb(0x8B,0x279); outb(0x45,0x279); outb(0xA2,0x279); outb(0xD1,0x279);
	outb(0xE8,0x279); outb(0x74,0x279); outb(0x3A,0x279); outb(0x9D,0x279); 
	outb(0xCE,0x279); outb(0xE7,0x279); outb(0x73,0x279); outb(0x39,0x279); 


	//wake[0] wake up the chip 
	outb(0x03,0x279);  //this will place the chip in isolation state
	outb(0x00,0xa79);   //wake[0]

	//Set the RD_DATA port
	outb(0x00,0x279);
	outb(0x80,0xa79);	 //read port address 203h

	//reading 72 bit serial identifier
 	outb(0x01,0x279);

	delay(100000); //1ms for hw memory access


//begin serial identifier, 72 bit read cycle
//	putstring("\nIdentifier (8 bytes): ");
	for(t=0;t<9;t++)
	{
		si[t]=Read_Byte();	//store the serial identifier in array
		delay(10000);		//250us delay necessary	
//		puthex(si[t]); putchar(' ');
        
	}  //end of isolation setup


	// Read the Checksum
	x=si[8]; 	//checksum field read

	//Assign and read the CSN.
 	outb(0x06,0x279);
	outb(0x02,0xa79);		// assign a card no = 2
        
	CSN=inb(0x203);
//	puthex(CSN);                     //check that it is assigned


	// Configuring logical devices.....
	//turn off logical devices 3 (modem) and 4 (cdrom)

	outb(0x07,0x279);
	outb(0x03,0xa79);
	outb(0x30,0x279);
	outb(0x00,0xa79);

	outb(0x07,0x279);
	outb(0x04,0xa79);
	outb(0x30,0x279);
	outb(0x00,0xa79);

//#define	MPU401
#ifndef MPU401
			//normally we do not need to enable MPU-401 (no use)
	outb(0x07,0x279);
	outb(0x02,0xa79);
	outb(0x30,0x279);
	outb(0x00,0xa79);

//...........................................configure logical device #2
#else
	outb(0x07,0x279);
	outb(0x02,0xa79);	//select logical device 2

	putstring("\nMPU-401 base: ");
	outb(0x61,0x279);	//set I/O base register (330)
	outb(0x30,0xa79);
	x=inb(0x203);		//check that the value was written

	outb(0x60,0x279);
	outb(0x03,0xa79);
	t=inb(0x203);
	puthex(x+(t<<8));

	putstring("\nMPU-401 IRQ: ");
	outb(0x70,0x279);	//write IRQL
	outb(0x03,0xa79);	//ask for IRQ3
	x=inb(0x203);
	puthex(x);

	outb(0x30,0x279);	//activate the logical device 
	outb(0x01,0xa79);
#endif
//...........................................configure logical device #5
#if 1
	outb(0x07,0x279);
	outb(0x05,0xa79);	//select logical device 5

	putstring("\nSlider base: ");
	outb(0x61,0x279);	//set I/O base register (201)
	outb(0x01,0xa79);
	x=inb(0x203);		//check that the value was written

	outb(0x60,0x279);
	outb(0x02,0xa79);
	t=inb(0x203);
	puthex(x+(t<<8));

	outb(0x30,0x279);	//activate the logical device 
	outb(0x01,0xa79);
#endif



//...........................................configure logical device #0
#if 1
	outb(0x07,0x279);
	outb(0x00,0xa79);	//select logical device 0

	putstring("\nWaveArtist base: ");

	outb(0x61,0x279);	//set I/O base register (250)
	outb(0x50,0xa79);
	x=inb(0x203);		//check that the value was written

	outb(0x60,0x279);
	outb(0x02,0xa79);
	t=inb(0x203);
	puthex(x+(t<<8));

	putstring("\nWA IRQ: ");
	outb(0x70,0x279);	//write IRQL
	outb(0x03,0xa79);	//ask for IRQ3
	x=inb(0x203);
	puthex(x);

	putstring("\nWA 16-bit DMA: ");
	outb(0x74,0x279);	//write DMA
	outb(0x07,0xa79);	//ask for DMA7
	x=inb(0x203);
	puthex(x);

	outb(0x30,0x279);	//activate the logical device 
	outb(0x01,0xa79);
#endif
//...........................................configure logical device #1

	outb(0x07,0x279);
	outb(0x01,0xa79);	//select logical device 1

	putstring("\nSound Blaster base: ");

	outb(0x61,0x279);	//set I/O base register (220)
	outb(0x20,0xa79);
	x=inb(0x203);		//check that the value was written

	outb(0x60,0x279);
	outb(0x02,0xa79);
	t=inb(0x203);
	puthex(x+(t<<8));

	putstring("\nSound Blaster IRQ: ");
	outb(0x70,0x279);	//write IRQL
	outb(0x03,0xa79);	//ask for IRQ3
	x=inb(0x203);
	puthex(x);

	putstring("\nSound Blaster 8-bit DMA: ");
	outb(0x74,0x279);	//write DMA
	outb(0x01,0xa79);	//ask for DMA1
	x=inb(0x203);
	puthex(x);

	putstring("\nAdLib base: ");

	outb(0x63,0x279);	//set I/O base register (388)
	outb(0x88,0xa79);
	x=inb(0x203);		//check that the value was written

	outb(0x62,0x279);
	outb(0x03,0xa79);
	t=inb(0x203);
	puthex(x+(t<<8));

	outb(0x30,0x279);	//activate the logical device 
	outb(0x01,0xa79);

//********************************************************************

	//After initialisation (reset), DSP returns AA..
	
	outb(0x01,0x226);	//DSP Reset sequence.
	delay(100000);		//need 3us delay
	outb(0x00,0x226);

	j=0x12345;
	while(!(inb(0x22e) & 0x80) && j)//poll the DSP read buffer status for valid data
	{
	    j--;
	
	}
	if (!j) putstring("\nSoundBlaster error: DSP reset failed!\n");
	
	x=inb(0x22a);			//read the data port

	putstring("\nSoundBlaster DSP reset (expect 0xAA): ");
	puthex(x);

	j=0x12345;
	while((inb(0x22c) & 0x80) && j)	//poll write buffer
	{
	    j--;
	
	}
	if (!j)
	{
	    putstring("\nSoundBlaster error: DSP not ready!\n");
	}
	else
	{
	
		outb(0xe1,0x22c);	//read DSP device ID

		putstring("\nSoundBlaster DSP id: ");
		x=inb(0x22a);		//SB+A
		delay(10);
		t=inb(0x22a);		//read lower and then upper byte
		puthex(x+(t<<8));

		//Turn off Speaker

		j=0x12345;
		while((inb(0x22c) & 0x80) && j)	//poll write buffer
		{
			j--;
	
		}
		
		if (!j)
		{
		        putstring("\nSoundBlaster error: Could not turn speaker off!\n");
		}
		else
		{	
			outb(0xd3,0x22c);

//			delay (1000000);	// need a delay of 112ms
//			while(inb(0x22c) & 0x80);

//			outb(0xd8,0x22c); 	//check speaker status
//			x=inb(0x22a);

//			putstring("\nSB Speaker OFF : ");
//      		hex(x);		//ff=On, 0x00=off


		drum();		//play start-up tune

		}
	}


	
//*********************************************************************

//do DMA BIOS-type init ...
#define DMA8_BASE	0x00
#define DMA16_BASE	0xC0

	outb(0xFF,DMA8_BASE+0x0D);	//master reset
	delay(100000);

//check 8-bit DMA regs...
	cTest = 0;
	for (cTest=0; cTest<255; cTest++)
	{
		t = DMA8_BASE;
		for (j=0;j<8;j++,t++)
		{
			outb(cTest,t);
			delay(25);
			outb(~cTest,t);	
			delay(25);

			x = inb(t);
			if (x != (cTest & 0xFF))	//error!!!
			{
				putstring("\nError verifying DMA8 reg #");
				puthex(t);
				putstring(", pattern ");
				puthex(cTest);
			}

			x = inb(t);
			if (x != ((~cTest) & 0xFF))	//error!!!
			{
				putstring("\nError verifying DMA8 reg #");
				puthex(t);
				putstring(", pattern [NOT] ");
				puthex(cTest);
			}
		}
	}

	outb(0xFF,DMA16_BASE+2*0x0D);	//master reset
	delay(25);

//check 16-bit DMA regs... Skip reg C0 and C2 - they do not exist...
	cTest = 0;
	for (cTest=0; cTest<255; cTest++)
	{
		t = DMA16_BASE+4;
		for (j=0;j<6;j++,t+=2)
		{
			outb(cTest,t);
			delay(25);
			outb(~cTest,t);	
			delay(25);

			x = inb(t);
			delay(25);
			if (x != (cTest & 0xFF))	//error!!!
			{
				putstring("\nError verifying DMA16 reg #");
				puthex(t);
				putstring(", pattern ");
				puthex(cTest);
			}

			x = inb(t);
			delay(25);
			if (x != ((~cTest) & 0xFF))	//error!!!
			{
				putstring("\nError verifying DMA16 reg #");
				puthex(t);
				putstring(", pattern [NOT] ");
				puthex(cTest);
			}
		}
	}


	disable_dma(0);
	disable_dma(1);
	disable_dma(2);
	disable_dma(3);
	disable_dma(5);
	disable_dma(6);
	disable_dma(7);

	outb(0x40,DMA8_BASE+0x0B);	//mode register,channel 0
	delay(25);
	outb(0x41,DMA8_BASE+0x0B);	//mode register,channel 1
	delay(25);
	outb(0x42,DMA8_BASE+0x0B);	//mode register,channel 2
	delay(25);
	outb(0x43,DMA8_BASE+0x0B);	//mode register,channel 3
	delay(25);

	outb(0xC0,DMA16_BASE+2*0x0B);	//mode register,channel 4 - cascade
	delay(25);
	outb(0x41,DMA16_BASE+2*0x0B);	//mode register,channel 5
	delay(25);
	outb(0x42,DMA16_BASE+2*0x0B);	//mode register,channel 6
	delay(25);
	outb(0x43,DMA16_BASE+2*0x0B);	//mode register,channel 7
	delay(25);
	enable_dma(4);

	//Woody 9.Sep.1998 change Priority Scheme from fixed to rotating
	outb(0x10,DMA8_BASE+8);	//DMA8 control reg - enable controller
	delay(25);
	outb(0x10,DMA16_BASE+2*8);	//DMA16 control reg - enable controller
	delay(25);
	
	
	//Woody 9.Sep.1998 Set all Extended Mode Register to DMA type F
	outb(0x30,0x40B);	//extended mode register, channel 0 type F
	delay(25);
	outb(0x31,0x40B);	//extended mode register, channel 1 type F
	delay(25);
	outb(0x32,0x40B);	//extended mode register, channel 2 type F
	delay(25);
	outb(0x33,0x40B);	//extended mode register, channel 3 type F
	delay(25);

	outb(0x31,0x4D6);	//extended mode register, channel 5 type F
	delay(25);
	outb(0x32,0x4D6);	//extended mode register, channel 6 type F
	delay(25);
	outb(0x33,0x4D6);	//extended mode register, channel 7 type F
	delay(25);

//end of BIOS type DMA init
//********************************************************************
/*
	*(volatile char*)0xE0500048 = 0x0F;	//enable ISA memory 00000-FFFFF
	putstring("\nBoard ID: ");
	x = *(volatile int *)0xD3000000;	// read a fake address 0 in ISA
	puthex(x);
*/

// enable all memory requests from ISA to be channeled to PCI
	*(volatile char*)0xE0500048 = 0xFF;

// for efficiency - we would like to set DMA to ping-pong-pack mode
//	*(volatile char*)0xE0500042 = 0x01;
// albeit as per errata - chip is buggy - disable ping-pong for now....
	*(volatile char*)0xE0500042 = 0x00;
	
// enable PCI packet retry
	*(volatile char*)0xE0500040 = 0x22;

// do not use PCI CPU park enable, park on last master, disable GAT bit
	*(volatile char*)0xE0500083 = 0x02;

	*(volatile char*)0xE0500080 = 0xE0;	//default rotataing priorities
	*(volatile char*)0xE0500081 = 0x01;	//rotate bank 4

// for Woody - disable all arbitration rotations between DMA, give
// the lowest proirity to the IDE channel...
// bk1 contains IDE, bk2 - VGA, bk3 - Eth100 and FB, bk0 - IDE and VidComp
//	*(volatile char*)0xE0500080 = 0x08;	//fixed bk1>bk2>bk3>bk0
//	*(volatile char*)0xE0500081 = 0x00;	//no Super Agent priority
//	*(volatile char*)0xE0500081 = 0x05;	//SIO Super Agent priority

//end of sound / DMA / 533 init
//********************************************************************

#ifdef XILINX_MODEM

	outb(0x84,0x338);	//RED LED on

//codec xilinx upload stuff...	
{
int i,j;
int c;
unsigned char* cPtr;
unsigned int count;

unsigned char* pName;
unsigned char* pDevice;
unsigned char* pDate;
unsigned char* pTime;
unsigned int unknown = 0x003F3F3F;	//'???'



#define VCSR_BASE	0xE1000000
#define	VFLASH_BASE	0xE1800000

//	putstring("\n *** Footbridge UART init ***\n");
	//start with stuffing some bytes to TX FIFO
	for (i=0; i<16; i++)
	{
		*(unsigned int *)(VCSR_BASE+UARTDR) = 0x00;
	}

	//set 8 bits, enable FIFO, 1 stop bit, no parity
	* (unsigned int *)(VCSR_BASE+H_UBRLCR) = 0x70;	//0 11 1 0 0 0 0

	//set baudrate to 19200 @66MHz (divisor = 225)
	* (unsigned int *)(VCSR_BASE+M_UBRLCR) = 0x00;
	* (unsigned int *)(VCSR_BASE+L_UBRLCR) = 0xE1;

	//flush the RX FIFO
	for (i=0; i<16; i++)
	{
		c = *(volatile unsigned int *)(VCSR_BASE+RXSTAT);
	}

	//enable UART
	* (unsigned int *)(VCSR_BASE+UARTCON) = 0x01;

	//Toggle the TX pin by seting BREAK ON, and then back OFF
	//set 8 bits, enable FIFO, 1 stop bit, no parity, BREAK ON
	* (unsigned int *)(VCSR_BASE+H_UBRLCR) = 0x71;	//0 11 1 0 0 0 1
	//set 8 bits, enable FIFO, 1 stop bit, no parity, BREAK OFF
	* (unsigned int *)(VCSR_BASE+H_UBRLCR) = 0x70;	//0 11 1 0 0 0 0

	//disable UART to set TX pin to HIGH
	* (unsigned int *)(VCSR_BASE+UARTCON) = 0x00;
	

//	putstring("\n *** Upload codec Xilinx. ***\n");


	//parse the header (I built this by looking at the data, not from docs)

	cPtr = (char*)(VFLASH_BASE+1008*1024);	//data is in last 16k of flash

	cPtr++;
	count = *(unsigned char*) cPtr;

	if (count == 0x09)
	{

	cPtr ++;	//advance past the halfword we just read
	//skip this field - non ASCII
	cPtr += count;
	cPtr += 4;	//unknown stuff


	count = *(unsigned char*)cPtr;
	cPtr++;		//advance past the halfword we just read
	if (count<20 && !*(cPtr+count-1))
		pName = cPtr;
	else
		pName = (unsigned char *)&unknown;
	cPtr += count;

	cPtr += 2;
	count = *(unsigned char*)cPtr;
	cPtr++;		//advance past the halfword we just read
	if (count<20 && !*(cPtr+count-1))
		pDevice = cPtr;
	else
		pDevice = (unsigned char *)&unknown;
	cPtr += count;


	cPtr += 2;
	count = *(unsigned char*) cPtr;
	cPtr++;		//advance past the halfword we just read
	if (count==9 && !*(cPtr+count-1))
		pDate = cPtr;
	else
		pDate = (unsigned char *)&unknown;
	cPtr += count;


	cPtr +=2;
	count = *(unsigned char*) cPtr;
	cPtr++;		//advance past the halfword we just read
	if (count==9 && !*(cPtr+count-1))
		pTime = cPtr;
	else
		pTime = (unsigned char *)&unknown;
	cPtr += count;

	cPtr += 3;
	count = *(unsigned char*)cPtr;
	count <<= 8;
	cPtr++;
	count += *(unsigned char*)cPtr;		//get size of bit array
	cPtr += 2;

	putstring("\nCodec name: ");
	putstring(pName);
	putstring(", device: ");
	putstring(pDevice);
	putstring(", rev.: ");
	putstring(pDate);
	putstring("  ");
	putstring(pTime);
	putstring("\n\n");


	for (i=0; i<count; i++)		//exactly 8838 bytes
	{
		c = *(char*)cPtr++;	//get next byte

		for(j=0; j<8;j++)
		{
			if (c & 0x80)
				outb(0x24,0x338);//GP15 bit set
			else
				outb(0x04,0x338);//GP15 bit clear

			outb(0x00,0x33A);	//clear clock
			outb(0x08,0x33A);	//set clock

			c<<=1;			//shift left by one
		}

//		for (j=0; j<1000; j++) {};

	}

	c = inb(0x338);				//read GP13
	if (c & 0x08)
		putstring("\nCodec programed OK.\n");
	else
		putstring("\nERROR: Codec not programmed OK...\n");

	}	//if count==9 - valid data
	else
	{
		putstring("\nERROR: Invalid codec data in flash memory...\n");

	}

}	//end of codec xilinx upload stuff...
	

#endif	/* XILINX_MODEM */



 	// !! more to do !!
	outb(0x06,0x338);	//green LED and FAN on...
	outb(0x00,0x33A);	//the default setting

}
/*************************************************************************/
//flag shared IO to be used by this process

int	getIOsemaphore(int processID)
{
unsigned long int	timeout;
int	bitset;

	//if this process already owns semaphore - all is OK
	if (processID == semaphoreID)
	    return(0);

	// do the resource arbitration first....
	timeout = jiffies;
	timeout += 500;	//give it a 5 second chance...
	bitset = 1;	//set default "busy" return

	while ((timeout > jiffies) && bitset)
	{
		if ((bitset = set_bit(31, &cShadowX)) != 0)
		{
			// if busy - wait till bit was not set...
			schedule();
		}
	}

	if (bitset)	//if busy all 5 seconds - error!
		return(-EBUSY);
		
	// else this process won the arbitration - put it's ID as the current ID
	semaphoreID = processID;
	return(0);
}

/*************************************************************************/
int	freeIOsemaphore(int processID)
{

	//if this process in not a semaphore owner - return error...
	if (processID != semaphoreID)
	    return(-EINVAL);
	    
	semaphoreID = 0;

//	if (clear_bit(31, &cShadowX) == 0)	//clear the lock bit		
//		return (-ENODEV);

//simply do it - it should be safe to unblock it...
	cShadowX &= 0x7FFFFFFF;

        return(0);
}

/*************************************************************************/
//toggle LED state between 0 = none, 1 = green, 2 = red, 3 = yellow

int	setLED(int state, int processID)
{
unsigned char	ctemp;
unsigned long	flags;

//	LEDs are fast - no need to semaphore this access
//	will be easier to CLI/STI around the process...

	save_flags(flags);
	cli();
	    
	ctemp = inb(0x338);
	ctemp &= ~(0x02+0x80);	//turn off both LEDs

	switch (state)
	{
		case 0:	break;
		case 1:	ctemp |= 0x02; break;	//green LED on
		case 2:	ctemp |= 0x80; break;	//red LED on
		case 3:	ctemp |= 0x82; break;	//both LEDs on
		default: restore_flags(flags); return (-EINVAL);
	}		
	outb(ctemp, 0x338);
	restore_flags(flags);

	return(state);
}

/*************************************************************************/
//return the current LED setting: 0 = none, 1 = green, 2 = red, 3 = yellow

int	getLED(void)
{
int	temp, ctemp;

		temp = 0;
		ctemp = inb(0x338);

		if (ctemp & 0x02) temp |= 1; //green LED bit
		if (ctemp & 0x80) temp |= 2; //red LED bit

		return(temp);
}

/*************************************************************************/
// The parameters are:
//	bitset - bit value to set/reset
//	bitmask	- bits to modify
//	e.g. to set bit 00001000: bitset=0x08, bitmask=0x08
//	     to clr bit 0000o000: bitset=0x00, bitmask=0x08
//return the new state of CPLD
int	set_CPLD(unsigned int bitset, unsigned int bitmask)
{
int i, shiftMask;
unsigned char current338;
unsigned char current33A;
int maskID;


	bitset &= 0x7FFFFFFF;		//do not permit to touch the HI bit!
	bitmask &= 0x7FFFFFFF;

	maskID = (bitset<<16)|(bitmask&0xFFFF);

	if (getIOsemaphore(maskID))
	    return(-EBUSY);

	//we want to write a bit pattern X110 to Xilinx, where X - therm_state
	//keep GREEN led on...
	//Bits: Thermostat reset/, Phillips7111 enable, Mute/, Flash Writable
	//The thermostat reset bit needs to be toggled to write data
	//IOCLOCK is bit 0x10, data is bit 0x20 on port 0x338
	//we added bit 0x40 as a flag to toggle red LED with CPU usage...
	
	current338 = inb(0x338) & 0xCF;	//clear 0x20 and 0x10
        current33A = inb(0x33A) & 0xFA;	//clear 0x04 and 0x01

	cShadowX &= ~bitmask;	//leave only bits not mentioned
	bitset &= bitmask;	//leave only important bits
	cShadowX |= bitset;	//merge the new bits with the old

	shiftMask = 0x08;	
	for (i=0; i<4; i++)
	{
		if (cShadowX & shiftMask)		//if bit was set
		{	
		    outb(current338 | 0x20, 0x338);	//data hi, clock low
		    outb(current338 | 0x30, 0x338);	//data hi, clock hi
		}
		else
		{	
		    outb(current338 | 0x00, 0x338);	//data low, clock low
		    outb(current338 | 0x10, 0x338);	//data low, clock hi
		}
		shiftMask >>= 1;
		
	}
	
	outb(current338, 0x338);	//data low, clock low

	//by now all 4 bits are shifted in. copy to IO reg, but leave clock HI
	outb(current33A | (0x01+0x04), 0x33A);
	outb(current33A | (0x00+0x04), 0x33A);	

	freeIOsemaphore(maskID);

	return(cShadowX & 0x7FFFFFFF);	//strip the potential LOCK bit
}
/*************************************************************************/
//and the sister function to read the current setting w/o exposing cShadowX
int	get_CPLD(void)
{
	return(cShadowX & 0x7FFFFFFF);	//strip the potential LOCK bit
}
/*************************************************************************/

void vnc_reset(void)
{
	/* set a yellow LED... */
	outb(0x86,0x338);

	/* shutdown sound... */
//	drum1();

 	/* open up the SuperIO chip */
 	outb(0x87,0x370);
 	outb(0x87,0x370);
	
 	outb(0x07,0x370);
 	outb(0x07,0x371);	// aux function group 1 (Logical Device 7)
	
 	outb(0xE6,0x370);
 	outb(0x00,0x371);	// set GP16 for (WD-TIMER) output

        /* set a RED LED and toggle WD_TIMER for rebooting... */
	outb(0xC4,0x338);

 	/* and return */
}
/*************************************************************************/

/* on rev 5 boards, the PCI bus can be tuned faster and slower */
void tune_up_pci_clock(void) {
    int i;
#ifdef DEBUG
    int k;
#endif
    volatile int *p = (volatile int*)(IO_BASE_ARM_CSR);
    int sample;
    int samples[16];

    /* turn on rtc for 3.90625 ms period*/
    outb(10, 0x70);
    outb(0x28, 0x71);
    
    /* set the access to bus tuning */
    p[SA_CONTROL/4] |= 0x70000000; /* enable cs outputs */
    p[XBUS_CYCLE/4] |= 7;          /* nice slow cycle */
    i = p[XBUS_STROBE/4];
    i &= ~0xFF;
    p[XBUS_STROBE/4] = i | 0xE1;     /* simple strobe pattern */

    /* create a time base, based on pci clock */
    p[TIMER_1_CONTROL/4] = 0xc0;     /* free run, reloading */
    p[TIMER_1_LOAD/4] = mSEC_10;     /* 66,666 */

    while (!(outb(12,0x70),inb(0x71) & 0x40));

    for (i=0; i<16; i++) {
        /* set bus speed; magic */
        *(volatile int*)(IO_BASE_DRAM_CONFIG
                         + (DC21285_XBUS_XCS0 - DC21285_DRAM_A0MR)
                         + 4*i) = 0;
        /* wait for start of period */
        while (!(outb(12,0x70),inb(0x71) & 0x40));
        /* sample timer 1 */
        sample = p[TIMER_1_VALUE/4];
        /* wait for end of period */
        while (!(outb(12,0x70),inb(0x71) & 0x40));
        sample -= p[TIMER_1_VALUE/4];
        if (sample<0) sample += p[TIMER_1_LOAD/4];
        if (sample>((mSEC_10 * 25)/64))
            break;
        //        printk("CLK_TUNE: bus speed %dKHz; increasing\n",(sample * 64)/250);
        samples[i] = sample;
    }
#ifdef DEBUG
    printk("CLK_TUNE: bus clk set to \n");
    for (k=0; k<i-1; k++)
        printk("%dKHz ",(samples[k] * 64)/250);
    printk("\nCLK_TUNE: final bus clk set to %dKHz\n",(sample * 64)/250);
#endif
}
